#!/usr/bin/env python
# coding: utf-8

# In[ ]:


from itertools import permutations
import numpy as np
import pandas as pd
    


# In[ ]:


def shortestpath(inputdataframe,distance):
    """
    Calculates shortest path from data matrix with symmetric TSP analytically
    
    Finds shortest path by iterating through all possible paths and keeping the shortest one.
    Uses the vector of indices of the data matrix as a path vector and permutates it, calculates the distance and compares to the previous value, then keeps the shorter one.
    
    Parameters
    ---------
    inpudataframe: Pandas DataFrame
                   Passed from salesman.salesman function (or use the salesman.dataimport function to retrieve it from file)
    distance: Float
              Maximum possible route length, passed from salesman.salesman
                   
    Returns
    ------
    shortest: Vector of strings
              Indices of inputdataframe in order of shortest route
    distance: Float
              Distance travelled by shortest route
    
    """
    pathvector = np.array(inputdataframe.index)
    permutating_vector = pathvector[:-1]
    
    for permutating_path in permutations(permutating_vector,len(permutating_vector)):
        path = np.append(permutating_path,pathvector[-1])
        rearrangeddata = inputdataframe.reindex(path,axis='columns')
        rolledpath = np.roll(path,len(path)-1)
        rearrangeddata = rearrangeddata.reindex(rolledpath,axis='index')
        pathlength = np.sum(np.diagonal(rearrangeddata))
        if pathlength < distance:
                distance = pathlength
                shortest = path 
                
    return distance, shortest
            

        




#!/usr/bin/env python
# coding: utf-8

# In[12]:


import numpy as np
import pandas as pd


# In[4]:


def nearestneighbour(inputdata,i,cityvector):
    """
    Finds a route by nearest neighbour algorithm for given starting point
    
    Finds the nearest still available (non-NaN) index in the matrix by searching for its column minimum, adds the index to the tour and the distance to tour length.
    Then replaces its own row by NaN and hands over to next stop it chose.
    
    Parameters
    ---------
    inputdata: Pandas DataFrame
               Matrix of symmetric TSP. Passed from findbestroute function
    i: integer
       Index position of starting point, passed from findbestroute function (iterator in findbestroute)
    cityvector: Vector of strings
                Vector containing the indices (nodes/"cities") of the data matrix, passed from findbestroute
    
    Returns
    ------
    route: Vector of strings
           Vector of indices in order of nearest neighbour route
    pathlengths: Float
                 Distance travelled by the nearest neighbour route
    """
    
    TSP = inputdata.copy()
    pathlength = 0
    node = cityvector[i]
    route = list([node])
    
    for j in range(len(cityvector)-1):
        pathlength += TSP[node].min()
        route.append(TSP[node].idxmin())
        TSP.loc[node,:] = np.nan
        node = route[j+1]
        j += 1
        
    return route, pathlength


# In[1]:


def findbestroute(inputdata,distance):
    """
    Iterates through different starting points for nearest neighbour and finds the shortest.
    
    Iterates through all indices of the data matrix and calls nearestneighbour on this starting point to find the nearest neighbour route, keeps the shortest.
    
    Parameters
    ---------
    inputdata: Pandas DataFrame
               Data matrix of the symmetric TSP, passed from salesman.salesman (or import from file with salesman.dataimport)
    distance: Float
              Longest possible distance (upper limit), passed from salesman.salesman
               
    Returns
    ------
    bestroute: Vector of strings
               Vector of data matrix indices in the order of shortest route it could find with this method
    distance:  Float
               Distance travelled by the output route
    """
    
    cityvector = np.array(inputdata.index)
    
    for i in range(len(cityvector)):
        route, pathlength = nearestneighbour(inputdata,i,cityvector)
        if pathlength < distance:
            distance = pathlength
            bestroute = route
        i += 1   
        
    return bestroute, distance
        


# In[ ]:





# In[ ]:







#!/usr/bin/env python
# coding: utf-8

# In[5]:


import numpy as np
import pandas as pd
from aspplea import exact
from aspplea import heuristic

# In[6]:


def dataimport(filename):
    """
    Imports data matrix from filename
    
    Parameters
    ----------
    filename: CSV file
              File should contain the (symmetric) data matrix with nodes/("cities") as column and row indices (as first row/column)
              
    Returns
    ------
    data: Pandas DataFrame
    
    """
    
    data = pd.read_csv(filename,sep=';',index_col=0,engine='python')
    return data


# In[7]:


def salesman(filename):
    """
    Finds best possible solution to the input symmetric TSP, either by exact calculation or a nearest-neighbour method
    
    Uses dataimport to import the data matrix, checks the size of the data matrix. 
    For problems with 8 or fewer nodes it calls the shortestpath function from exact to calculate exact solution.
    For problems with more than 8 nodes it calls the findbestroute function from heuristic to calculate the best possible solution available to the nearest neighbour algorithm.
    In the latter case, it calculates the lower and upper limit and mean path and prints them for comparison.
    
    Parameters
    ---------
    filename: CSV file 
              See dataimport
              
    Returns
    ------
    shortest: Vector or list of strings (nodes/cities)
              Indices of data in order of shortest route
              Either absolute shortest route or shortest route it could find for problems greater 8
    distance: Float
              Distance travelled by shortest route
    """
    data = dataimport(filename)
    Maxlength = 0
    for i in range(len(data)):
        Maxlength +=data.iloc[:,i].max()
    
    if data.size <= 64:
        print("Exact method will be used")
        #Load exact method
        distance, shortest = exact.shortestpath(data,Maxlength)
        print("Shortest path: ", shortest)
        print("Length: ", distance)
        
    else:
        print("I will use a heuristic method to find a best guess as the dataset is larger than 8")
        shortest, distance = heuristic.findbestroute(data,Maxlength)
        print("This is the shortest route I could find with the nearest neighbour method: ", shortest)
        print("Length: ", distance)
    
        print("Here is how it compares:")
        Minlength = 0
        for i in range(len(data)):
            Minlength +=data.iloc[:,i].min()
        print("Sum of distances from each city to nearest neighbour: ",Minlength)
        print("Sum of distances from each city to farthest neighbour: ",Maxlength)
        Average = len(data)*data.mean().mean()
        print("Average distance of random tour: ",Average)
        
    return shortest,distance


# In[8]:





# In[ ]:





# In[ ]:





# In[ ]:




